#ifndef TESTTEST_H
#define TESTTEST_H

#include "stdafx.h"
#include "..\..\Shared\pkcs11Stub.h"

#include "cxxtest/TestSuite.h"

#ifdef CRYPTOAMOR64
  #define VIRTPKCS11_DLL   _T("VirtPKCS1164.dll")
#else
  #define VIRTPKCS11_DLL   _T("VirtPKCS11.dll")
#endif

//#define DISABLE_SLOW_TESTS

// INFO ABOUT AVAILABLE READERS 
//
#define MAX_READER_NAME_LENGTH  256
#define TESTING_TOKEN_LABEL		 "CxxTest"

typedef struct _ITEM_DATA_PKCS11READER {
    ULONG               slotID;
    ULONG               hSession;
    TCHAR               readerName[MAX_READER_NAME_LENGTH];
    BOOL                bCardPresent;
    TCHAR               cardATR[MAX_READER_NAME_LENGTH];

    _ITEM_DATA_PKCS11READER() {
        clear();
    }

    void clear() {
        slotID = 0;
        hSession = NULL;
        memset(readerName, 0, sizeof(readerName));
        bCardPresent = FALSE;
        memset(cardATR, 0, sizeof(cardATR));
    }
} ITEM_DATA_PKCS11READER;

typedef list<ITEM_DATA_PKCS11READER> lp11r;


class MyTestSuite : public CxxTest::TestSuite 
{
    CPKCS11Stub m_pkcs11Mngr;     // MANAGER FOR PKCS#11 TOKENS

	int getTestTokenSession(CK_SLOT_ID* pSlotID, CK_SLOT_INFO* pSlotInfo, CK_TOKEN_INFO* pTokenInfo) {
		int             status = STAT_OK;
		CK_SLOT_ID      pkcs11Slots[100];      
		DWORD           pkcs11SlotsCount = 100;      
		ITEM_DATA_PKCS11READER  readerInfo;
		m_pkcs11Mngr.Init(VIRTPKCS11_DLL, TRUE);

		TS_ASSERT(m_pkcs11Mngr.C_GetSlotList(FALSE, pkcs11Slots, &pkcs11SlotsCount) == CKR_OK);
		TS_ASSERT(pkcs11SlotsCount > 0);
		BOOL	bTestTokenFound = FALSE;
		for (DWORD i = 0; i < pkcs11SlotsCount; i++) {
			TS_ASSERT(m_pkcs11Mngr.C_GetSlotInfo(pkcs11Slots[i], pSlotInfo) == CKR_OK);
			TS_ASSERT(m_pkcs11Mngr.C_GetTokenInfo(pkcs11Slots[i], pTokenInfo) == CKR_OK);
					
			// Do only when testing token was found
			if (strcmp((const char*)pTokenInfo->label, TESTING_TOKEN_LABEL) == 0) {
				bTestTokenFound = TRUE;
				*pSlotID = pkcs11Slots[i];
				break;
			}
		}

		if (!bTestTokenFound) status = CKR_TOKEN_NOT_PRESENT;

		return status;
	}

public:
	
	void testCreateEmptyToken(void) {
/*		
		int             status = STAT_OK;
		HINSTANCE hLib = LoadLibrary(VIRTPKCS11_DLL);
		if (hLib) {
			FT_CAM_CreateBlankToken	functPtr = NULL;
			if ((functPtr = (FT_CAM_CreateBlankToken) GetProcAddress(hLib, "CAM_CreateBlankToken")) != NULL) {
				TS_ASSERT((status = (functPtr)((CK_CHAR_PTR) "D:\\VirtToken_empty.vtk", (CK_CHAR_PTR) TESTING_TOKEN_LABEL, (CK_CHAR_PTR) "12345678", TRUE)) == CKR_OK);
			}
			FreeLibrary(hLib);
		}
/**/
	}

    void testListReaders( void ) {
	// PKCS#11 READERS
		int             status = STAT_OK;
		CK_SLOT_ID      pkcs11Slots[100];      
		DWORD           pkcs11SlotsCount = 100;      
		CK_SLOT_INFO    slotInfo;
		CK_TOKEN_INFO   tokenInfo;
		ITEM_DATA_PKCS11READER  readerInfo;
		m_pkcs11Mngr.Init(VIRTPKCS11_DLL);

		if ((status = m_pkcs11Mngr.C_GetSlotList(FALSE, pkcs11Slots, &pkcs11SlotsCount)) == STAT_OK) {
			TS_ASSERT(pkcs11SlotsCount > 0);
			for (DWORD i = 0; i < pkcs11SlotsCount; i++) {
				if (m_pkcs11Mngr.C_GetSlotInfo(pkcs11Slots[i], &slotInfo) == STAT_OK) {
					if (m_pkcs11Mngr.C_GetTokenInfo(pkcs11Slots[i], &tokenInfo) == STAT_OK) {
						readerInfo.clear(); 
						readerInfo.slotID = pkcs11Slots[i];
	    
						readerInfo.bCardPresent = TRUE;
						tokenInfo.label[31] = 0x00;
						slotInfo.slotDescription[63] = 0x00;
						_tcscpy_s(readerInfo.cardATR, MAX_READER_NAME_LENGTH, (TCHAR*) tokenInfo.label);
						_tcscpy_s(readerInfo.readerName, MAX_READER_NAME_LENGTH, (TCHAR*) slotInfo.slotDescription);
	                    
						// IF MORE THAN 3 SPACES DETECTED, THAN CUT READER NAME
						TCHAR* pos = 0;
						if ((pos = _tcsstr(readerInfo.cardATR, _T("   "))) != NULL) {
							memset(readerInfo.cardATR + (pos - readerInfo.cardATR) * sizeof(TCHAR), 0, sizeof(TCHAR));
						}
						if ((pos = _tcsstr(readerInfo.readerName, _T("   "))) != NULL) {
							memset(readerInfo.readerName + (pos - readerInfo.readerName) * sizeof(TCHAR), 0, sizeof(TCHAR));
						}
					}
					else  TS_FAIL("C_GetTokenInfo() fail");
				}
				else  TS_FAIL("C_GetSlotInfo fail");
			}
		}
		else  TS_FAIL("C_GetSlotList fail");
	}


    void testTokenPINTests( void ) {
#ifdef DISABLE_SLOW_TESTS
		TS_FAIL("testTokenPINTests disabled");
#else
		// PKCS#11 READERS
		int             status = STAT_OK;

		CK_SLOT_ID	slotID = -100;
		CK_SLOT_INFO    slotInfo;
		CK_TOKEN_INFO   tokenInfo;
		TS_ASSERT((status = getTestTokenSession(&slotID, &slotInfo, &tokenInfo)) == CKR_OK);
		if (status == CKR_OK) {
				#define TEST_PIN	"123456"
				#define TEST_PIN2	"123488"
				#define DEFAULT_ADMIN_PIN			"12345678"
				#define WRONG_ADMIN_PIN				"12345679"
				

				DeleteFile((LPCSTR) slotInfo.slotDescription);
				HINSTANCE hLib = LoadLibrary(VIRTPKCS11_DLL);
				if (hLib) {
					FT_CAM_CreateBlankToken	functPtr = NULL;
					if ((functPtr = (FT_CAM_CreateBlankToken) GetProcAddress(hLib, "CAM_CreateBlankToken")) != NULL) {
						status = (functPtr)((CK_CHAR_PTR) slotInfo.slotDescription, (CK_CHAR_PTR) TESTING_TOKEN_LABEL, (CK_CHAR_PTR) DEFAULT_ADMIN_PIN, TRUE);
					}
					FreeLibrary(hLib);
				}
				CK_SESSION_HANDLE   hSession;
				// Init token 
				TS_ASSERT(m_pkcs11Mngr.C_InitToken(slotID, (CK_CHAR*)TEST_PIN, strlen(TEST_PIN), tokenInfo.label) == CKR_OK);
				// User pin not initialized yet, login should fail
				TS_ASSERT(m_pkcs11Mngr.C_OpenSession(slotID, CKF_RW_SESSION | CKF_SERIAL_SESSION, NULL_PTR, NULL_PTR, &hSession) == CKR_OK);
				TS_ASSERT(m_pkcs11Mngr.C_Login(hSession, CKU_USER, (CK_CHAR*)TEST_PIN, strlen(TEST_PIN)) == CKR_USER_PIN_NOT_INITIALIZED);

				// Set User PIN
				TS_ASSERT(m_pkcs11Mngr.C_OpenSession(slotID, CKF_RW_SESSION | CKF_SERIAL_SESSION, NULL_PTR, NULL_PTR, &hSession) == CKR_OK);
				TS_ASSERT(m_pkcs11Mngr.C_Login(hSession, CKU_SO, (CK_CHAR*)DEFAULT_ADMIN_PIN, strlen(DEFAULT_ADMIN_PIN)) == CKR_OK);
				// Logging again should fail - but disabled, as some apps are failing with that
				// TS_ASSERT(m_pkcs11Mngr.C_Login(hSession, CKU_SO, (CK_CHAR*)DEFAULT_ADMIN_PIN, strlen(DEFAULT_ADMIN_PIN)) == CKR_USER_ALREADY_LOGGED_IN);
				TS_ASSERT(m_pkcs11Mngr.C_InitPIN(hSession, (CK_CHAR*)TEST_PIN, strlen(TEST_PIN)) == CKR_OK);

				// Login user
				TS_ASSERT(m_pkcs11Mngr.C_Login(hSession, CKU_USER, (CK_CHAR*)TEST_PIN, strlen(TEST_PIN)) == CKR_OK);
				// Logging again should fail - but disabled, as some apps are failing with that
				//TS_ASSERT(m_pkcs11Mngr.C_Login(hSession, CKU_USER, (CK_CHAR*)TEST_PIN, strlen(TEST_PIN)) == CKR_USER_ALREADY_LOGGED_IN);
				TS_ASSERT(m_pkcs11Mngr.C_Logout(hSession) == CKR_OK);
				// Wrong user PIN
				// Four wrong PINS should lock the PIN
				for (int numRep = 0; numRep < 5; numRep++) TS_ASSERT(m_pkcs11Mngr.C_Login(hSession, CKU_USER, (CK_CHAR*)TEST_PIN2, strlen(TEST_PIN2)) == CKR_PIN_INVALID);
				TS_ASSERT(m_pkcs11Mngr.C_Login(hSession, CKU_USER, (CK_CHAR*)TEST_PIN2, strlen(TEST_PIN2)) == CKR_PIN_LOCKED);
				TS_ASSERT(m_pkcs11Mngr.C_Login(hSession, CKU_USER, (CK_CHAR*)TEST_PIN, strlen(TEST_PIN)) == CKR_PIN_LOCKED);
				
				// Log admin and then try to change admin PIN with user PIN value (should fail)
				TS_ASSERT(m_pkcs11Mngr.C_Login(hSession, CKU_SO, (CK_CHAR*)DEFAULT_ADMIN_PIN, strlen(DEFAULT_ADMIN_PIN)) == CKR_OK);
				TS_ASSERT(m_pkcs11Mngr.C_SetPIN(hSession, (CK_CHAR*)TEST_PIN, strlen(TEST_PIN), (CK_CHAR*)TEST_PIN2, strlen(TEST_PIN2)) == CKR_PIN_INVALID);
				// User PIN should be still locked
				TS_ASSERT(m_pkcs11Mngr.C_Login(hSession, CKU_USER, (CK_CHAR*)TEST_PIN2, strlen(TEST_PIN2)) == CKR_PIN_LOCKED);

				// Login admin
				TS_ASSERT(m_pkcs11Mngr.C_Login(hSession, CKU_SO, (CK_CHAR*)DEFAULT_ADMIN_PIN, strlen(DEFAULT_ADMIN_PIN)) == CKR_OK);
				TS_ASSERT(m_pkcs11Mngr.C_InitPIN(hSession, (CK_CHAR*)TEST_PIN, strlen(TEST_PIN)) == CKR_OK);
				TS_ASSERT(m_pkcs11Mngr.C_Login(hSession, CKU_USER, (CK_CHAR*)TEST_PIN, strlen(TEST_PIN)) == CKR_OK);

				// Login admin
				TS_ASSERT(m_pkcs11Mngr.C_Login(hSession, CKU_SO, (CK_CHAR*)DEFAULT_ADMIN_PIN, strlen(DEFAULT_ADMIN_PIN)) == CKR_OK);
				TS_ASSERT(m_pkcs11Mngr.C_Logout(hSession) == CKR_OK);

				// Login admin with wrong pi and try to init user pin
				TS_ASSERT(m_pkcs11Mngr.C_Login(hSession, CKU_SO, (CK_CHAR*)WRONG_ADMIN_PIN, strlen(WRONG_ADMIN_PIN)) == CKR_PIN_INVALID);
				TS_ASSERT(m_pkcs11Mngr.C_InitPIN(hSession, (CK_CHAR*)TEST_PIN, strlen(TEST_PIN)) == CKR_USER_NOT_LOGGED_IN);
				TS_ASSERT(m_pkcs11Mngr.C_CloseAllSessions(slotID) == CKR_OK);
		}
/**/
#endif
	}

    void testInsertObjects( void ) {
		// PKCS#11 READERS
		int             status = STAT_OK;

		CK_SLOT_ID	slotID = -100;
		CK_SLOT_INFO    slotInfo;
		CK_TOKEN_INFO   tokenInfo;
		TS_ASSERT((status = getTestTokenSession(&slotID, &slotInfo, &tokenInfo)) == CKR_OK);
		if (status == CKR_OK) {

				#define TEST_PIN	"123456"
#ifndef DEFAULT_ADMIN_PIN
				#define DEFAULT_ADMIN_PIN			"12345678"
#endif

				#define NUM_OBJECTS		10
				CK_OBJECT_HANDLE hObject[NUM_OBJECTS];
				CK_ULONG ulObjectCount;
				// TEMPLATE to use for creation/searching object on token
				CK_CHAR label_private[] = {"Test1_private"};	//label of data object
				CK_OBJECT_CLASS dataClass = CKO_DATA;  //type of data in object
				CK_BBOOL ptrue = TRUE; CK_BBOOL pfalse = FALSE; 
				CK_CHAR	data_private[] = {"CxxTest Private"};
				CK_ATTRIBUTE dataTemplate_private[] = {
					{CKA_CLASS, &dataClass, sizeof(dataClass)},
					{CKA_TOKEN, &ptrue, sizeof(ptrue)},
					{CKA_LABEL, label_private, sizeof(label_private)},
					{CKA_VALUE, (CK_VOID_PTR) data_private, sizeof(data_private)},
					{CKA_PRIVATE, &ptrue, sizeof(ptrue)}  // private object
				};
				BYTE    numAttributes_private = 5;
				
				CK_CHAR label_public[] = {"Test1_public"};	//label of data object
				CK_CHAR	data_public[] = {"CxxTest Public"};
				CK_ATTRIBUTE dataTemplate_public[] = {
					{CKA_CLASS, &dataClass, sizeof(dataClass)},
					{CKA_TOKEN, &ptrue, sizeof(ptrue)},
					{CKA_LABEL, label_public, sizeof(label_public)},
					{CKA_VALUE, (CK_VOID_PTR) data_public, sizeof(data_public)},
					{CKA_PRIVATE, &pfalse, sizeof(pfalse)}  // private object
				};
				BYTE    numAttributes_public = 5;


				CK_SESSION_HANDLE	hSession;
				TS_ASSERT(m_pkcs11Mngr.C_OpenSession(slotID, CKF_RW_SESSION | CKF_SERIAL_SESSION, NULL_PTR, NULL_PTR, &hSession) == CKR_OK);

				// Create new private object - should fail as we are not logged in 
				TS_ASSERT(m_pkcs11Mngr.C_CreateObject(hSession, dataTemplate_private, numAttributes_private, &hObject[0]) != CKR_OK);

				// Create new public object - should success even when we are not logged in 
				TS_ASSERT(m_pkcs11Mngr.C_CreateObject(hSession, dataTemplate_public, numAttributes_public, &hObject[0]) == CKR_OK);


				// Login user
				TS_ASSERT(m_pkcs11Mngr.C_Login(hSession, CKU_USER, (CK_CHAR*)TEST_PIN, strlen(TEST_PIN)) == CKR_OK);

				// Create new object - should succedd
				TS_ASSERT(m_pkcs11Mngr.C_CreateObject(hSession, dataTemplate_private, numAttributes_private, &hObject[0]) == CKR_OK);

				// Login SO
				TS_ASSERT(m_pkcs11Mngr.C_Login(hSession, CKU_SO, (CK_CHAR*)DEFAULT_ADMIN_PIN, strlen(DEFAULT_ADMIN_PIN)) == CKR_OK);
				// Create new private object - should succed
				((char*) dataTemplate_private[2].pValue)[2] = 'x';	// changle character in label to create unused label
				TS_ASSERT(m_pkcs11Mngr.C_CreateObject(hSession, dataTemplate_private, numAttributes_private, &hObject[0]) == CKR_OK);

				// TODO:Create same object again - should fail??
				//TS_ASSERT(m_pkcs11Mngr.C_CreateObject(hSession, dataTemplate_private, numAttributes_private, &hObject[0]) != CKR_OK);

				// Logout	
				TS_ASSERT(m_pkcs11Mngr.C_Logout(hSession) == CKR_OK);


				// No private objects should be found as we are logged out and object is private
				TS_ASSERT(m_pkcs11Mngr.C_FindObjectsInit(hSession, dataTemplate_private, numAttributes_private) == CKR_OK); 
				TS_ASSERT(m_pkcs11Mngr.C_FindObjects(hSession, hObject, NUM_OBJECTS, &ulObjectCount) == CKR_OK);	
				TS_ASSERT(m_pkcs11Mngr.C_FindObjectsFinal(hSession) == CKR_OK); 
				TS_ASSERT(ulObjectCount == 0);

				// One public object should be found even when we are logged out
				TS_ASSERT(m_pkcs11Mngr.C_FindObjectsInit(hSession, dataTemplate_public, numAttributes_public) == CKR_OK); 
				TS_ASSERT(m_pkcs11Mngr.C_FindObjects(hSession, hObject, NUM_OBJECTS, &ulObjectCount) == CKR_OK);	
				TS_ASSERT(m_pkcs11Mngr.C_FindObjectsFinal(hSession) == CKR_OK); 
				TS_ASSERT(ulObjectCount == 1);

				// Partial template searching - one public object should be found (CKA_CLASS, CKA_TOKEN, CKA_LABEL)
				TS_ASSERT(m_pkcs11Mngr.C_FindObjectsInit(hSession, dataTemplate_public, 3) == CKR_OK); 
				TS_ASSERT(m_pkcs11Mngr.C_FindObjects(hSession, hObject, NUM_OBJECTS, &ulObjectCount) == CKR_OK);	
				TS_ASSERT(m_pkcs11Mngr.C_FindObjectsFinal(hSession) == CKR_OK); 
				TS_ASSERT(ulObjectCount == 1);

				// Partial template searching - one public object should be found (CKA_CLASS, CKA_TOKEN)
				TS_ASSERT(m_pkcs11Mngr.C_FindObjectsInit(hSession, dataTemplate_public, 2) == CKR_OK); 
				TS_ASSERT(m_pkcs11Mngr.C_FindObjects(hSession, hObject, NUM_OBJECTS, &ulObjectCount) == CKR_OK);	
				TS_ASSERT(m_pkcs11Mngr.C_FindObjectsFinal(hSession) == CKR_OK); 
				TS_ASSERT(ulObjectCount == 1);

				// Login user
				TS_ASSERT(m_pkcs11Mngr.C_Login(hSession, CKU_USER, (CK_CHAR*)TEST_PIN, strlen(TEST_PIN)) == CKR_OK);
				TS_ASSERT(m_pkcs11Mngr.C_FindObjectsInit(hSession, dataTemplate_private, numAttributes_private) == CKR_OK); 
				TS_ASSERT(m_pkcs11Mngr.C_FindObjects(hSession, hObject, NUM_OBJECTS, &ulObjectCount) == CKR_OK);	
				TS_ASSERT(m_pkcs11Mngr.C_FindObjectsFinal(hSession) == CKR_OK); 
				// Only one object should be present
				TS_ASSERT(ulObjectCount == 1);
				CK_ATTRIBUTE dataTemplate[] = { {CKA_VALUE, NULL_PTR, 0},};
				int numAttributes = 1;

				// Partial template searching - one public and one private object should be found (CKA_CLASS, CKA_TOKEN)
				TS_ASSERT(m_pkcs11Mngr.C_FindObjectsInit(hSession, dataTemplate_public, 2) == CKR_OK); 
				TS_ASSERT(m_pkcs11Mngr.C_FindObjects(hSession, hObject, NUM_OBJECTS, &ulObjectCount) == CKR_OK);	
				TS_ASSERT(m_pkcs11Mngr.C_FindObjectsFinal(hSession) == CKR_OK); 
				TS_ASSERT(ulObjectCount == 3);

				// Find private object
				TS_ASSERT(m_pkcs11Mngr.C_FindObjectsInit(hSession, dataTemplate_private, numAttributes_private) == CKR_OK); 
				TS_ASSERT(m_pkcs11Mngr.C_FindObjects(hSession, hObject, NUM_OBJECTS, &ulObjectCount) == CKR_OK);	
				TS_ASSERT(m_pkcs11Mngr.C_FindObjectsFinal(hSession) == CKR_OK); 
				TS_ASSERT(ulObjectCount == 1);

				// LOAD 
				TS_ASSERT(m_pkcs11Mngr.C_GetAttributeValue(hSession, hObject[0], dataTemplate, numAttributes) == CKR_OK);
				dataTemplate[0].pValue = new char[dataTemplate[0].ulValueLen + 1];
				memset(dataTemplate[0].pValue, 0, dataTemplate[0].ulValueLen + 1);
				TS_ASSERT(m_pkcs11Mngr.C_GetAttributeValue(hSession, hObject[0], dataTemplate, numAttributes) == CKR_OK);
				TS_ASSERT(dataTemplate[0].ulValueLen == sizeof(data_private));
				TS_ASSERT(memcmp(dataTemplate[0].pValue, data_private, sizeof(data_private)) == 0);
				delete[] dataTemplate[0].pValue;

				TS_ASSERT(m_pkcs11Mngr.C_DestroyObject(hSession, hObject[0]) == CKR_OK);
				TS_ASSERT(m_pkcs11Mngr.C_CloseAllSessions(slotID) == CKR_OK);
		/**/
		}
	}

    void testRandom(void) {
		int             status = STAT_OK;

		CK_SLOT_ID	slotID = -100;
		CK_SLOT_INFO    slotInfo;
		CK_TOKEN_INFO   tokenInfo;
		TS_ASSERT((status = getTestTokenSession(&slotID, &slotInfo, &tokenInfo)) == CKR_OK);
		if (status == CKR_OK) {
				CK_SESSION_HANDLE	hSession;
				TS_ASSERT(m_pkcs11Mngr.C_OpenSession(slotID, CKF_RW_SESSION | CKF_SERIAL_SESSION, NULL_PTR, NULL_PTR, &hSession) == CKR_OK);

				// Small chunk of random data
				BYTE	random1[10];
				TS_ASSERT(m_pkcs11Mngr.C_GenerateRandom(hSession, random1, sizeof(random1)) == CKR_OK);
				BYTE	random2[10];
				TS_ASSERT(m_pkcs11Mngr.C_GenerateRandom(hSession, random2, sizeof(random2)) == CKR_OK);
				DWORD	dwDiffer = 0;
				for (int i = 0; i < sizeof(random1); i++) if (random1[i] != random2[i]) dwDiffer++;
				TS_ASSERT(((float) dwDiffer / sizeof(random1)) > 0.9);		// 9 out of 10 must be different
	
				// Big chunk of random data
				BYTE	random3[1000];
				TS_ASSERT(m_pkcs11Mngr.C_GenerateRandom(hSession, random3, sizeof(random3)) == CKR_OK);
				BYTE	random4[1000];
				TS_ASSERT(m_pkcs11Mngr.C_GenerateRandom(hSession, random4, sizeof(random4)) == CKR_OK);
				dwDiffer = 0;
				for (int i = 0; i < sizeof(random3); i++) if (random3[i] != random4[i]) dwDiffer++;
				TS_ASSERT(((float) dwDiffer / sizeof(random3)) > 0.95);		// 90% of bytes must be different
		}
		TS_ASSERT(m_pkcs11Mngr.C_CloseAllSessions(slotID) == CKR_OK);
/**/
	}

    void testEncryptedUserObjects(void) {
		int             status = STAT_OK;

		CK_SLOT_ID	slotID = -100;
		CK_SLOT_INFO    slotInfo;
		CK_TOKEN_INFO   tokenInfo;
		TS_ASSERT((status = getTestTokenSession(&slotID, &slotInfo, &tokenInfo)) == CKR_OK);
		if (status == CKR_OK) {
			CK_SESSION_HANDLE	hSession;

			CK_OBJECT_HANDLE hObject[NUM_OBJECTS];
			// TEMPLATE to use for creation/searching object on token
			CK_CHAR label_private[] = {"Test2_private_AAAAAAAA"};	//label of data object
			CK_OBJECT_CLASS dataClass = CKO_DATA;  //type of data in object
			CK_BBOOL ptrue = TRUE; CK_BBOOL pfalse = FALSE; 
			CK_CHAR	data_private[] = {"CxxTest Private_BBBBBBBB"};
			CK_ATTRIBUTE dataTemplate_private[] = {
				{CKA_CLASS, &dataClass, sizeof(dataClass)},
				{CKA_TOKEN, &ptrue, sizeof(ptrue)},
				{CKA_LABEL, label_private, sizeof(label_private)},
				{CKA_VALUE, (CK_VOID_PTR) data_private, sizeof(data_private)},
				{CKA_PRIVATE, &ptrue, sizeof(ptrue)}  // private object
			};
			BYTE    numAttributes_private = 5;
			CK_CHAR label_public[] = {"Test1_public_CCCCCCCC"};	//label of data object
			CK_CHAR	data_public[] = {"CxxTest Public_DDDDDDDD"};
			CK_ATTRIBUTE dataTemplate_public[] = {
				{CKA_CLASS, &dataClass, sizeof(dataClass)},
				{CKA_TOKEN, &ptrue, sizeof(ptrue)},
				{CKA_LABEL, label_public, sizeof(label_public)},
				{CKA_VALUE, (CK_VOID_PTR) data_public, sizeof(data_public)},
				{CKA_PRIVATE, &pfalse, sizeof(pfalse)}  // private object
			};
			BYTE    numAttributes_public = 5;

			TS_ASSERT(m_pkcs11Mngr.C_OpenSession(slotID, CKF_RW_SESSION | CKF_SERIAL_SESSION, NULL_PTR, NULL_PTR, &hSession) == CKR_OK);

			// Create new public object 
			TS_ASSERT(m_pkcs11Mngr.C_CreateObject(hSession, dataTemplate_public, numAttributes_public, &hObject[0]) == CKR_OK);
			// Login user
			TS_ASSERT(m_pkcs11Mngr.C_Login(hSession, CKU_USER, (CK_CHAR*)TEST_PIN, strlen(TEST_PIN)) == CKR_OK);
			// Create new private object - should succedd
			TS_ASSERT(m_pkcs11Mngr.C_CreateObject(hSession, dataTemplate_private, numAttributes_private, &hObject[0]) == CKR_OK);
			// Close session
			TS_ASSERT(m_pkcs11Mngr.C_CloseSession(hSession) == CKR_OK);


			// Open token file on raw level
			CFile	file;
			if (file.Open((LPCTSTR) slotInfo.slotDescription, CFile::modeRead)) {
				CString	tokenContent;
				if (file.Read(tokenContent.GetBuffer((int) file.GetLength() + 100),  (int)(file.GetLength() + 100)) > 0) {
					tokenContent.ReleaseBuffer();

					// CCCCCCCC and DDDDDDDD should *be* present in token content (public, non-encrypted part)
					TS_ASSERT(tokenContent.Find("43 43 43 43 43 43 43 43") != -1);
					TS_ASSERT(tokenContent.Find("44 44 44 44 44 44 44 44") != -1);

					// AAAAAAAA and BBBBBBBB should *not* be present in clear text in token content (private, encrypted part)
					TS_ASSERT(tokenContent.Find("41 41 41 41 41 41 41 41") == -1);
					TS_ASSERT(tokenContent.Find("42 42 42 42 42 42 42 42") == -1);
					TS_ASSERT(tokenContent.Find("AAAAAAAA") == -1);
				}
				else {
					TS_FAIL("testEncryptedUserObjects: fail to read test token");
				}

				file.Close();
			}
			else {
				TS_FAIL("testEncryptedUserObjects: fail to open test token");
			}
		}
/**/
	}
    void testMultipleSessionCacheIndependence(void) {
		int             status = STAT_OK;

		CK_SLOT_ID	slotID = -100;
		CK_SLOT_INFO    slotInfo;
		CK_TOKEN_INFO   tokenInfo;
		TS_ASSERT((status = getTestTokenSession(&slotID, &slotInfo, &tokenInfo)) == CKR_OK);
		if (status == CKR_OK) {
			CK_SESSION_HANDLE	hPublicSession;
			CK_SESSION_HANDLE	hPrivateSession;

			CK_OBJECT_HANDLE hObject[NUM_OBJECTS];
			CK_ULONG ulObjectCount;
			// TEMPLATE to use for creation/searching object on token
			CK_CHAR label_private[] = {"Test3_private"};	//label of data object
			CK_OBJECT_CLASS dataClass = CKO_DATA;  //type of data in object
			CK_BBOOL ptrue = TRUE; CK_BBOOL pfalse = FALSE; 
			CK_CHAR	data_private[] = {"CxxTest Test3_private"};
			CK_ATTRIBUTE dataTemplate_private[] = {
				{CKA_CLASS, &dataClass, sizeof(dataClass)},
				{CKA_TOKEN, &ptrue, sizeof(ptrue)},
				{CKA_LABEL, label_private, sizeof(label_private)},
				{CKA_VALUE, (CK_VOID_PTR) data_private, sizeof(data_private)},
				{CKA_PRIVATE, &ptrue, sizeof(ptrue)}  // private object
			};
			BYTE    numAttributes_private = 5;
			CK_CHAR label_public[] = {"Test3_public"};	//label of data object
			CK_CHAR	data_public[] = {"CxxTest Test3_public"};
			CK_ATTRIBUTE dataTemplate_public[] = {
				{CKA_CLASS, &dataClass, sizeof(dataClass)},
				{CKA_TOKEN, &ptrue, sizeof(ptrue)},
				{CKA_LABEL, label_public, sizeof(label_public)},
				{CKA_VALUE, (CK_VOID_PTR) data_public, sizeof(data_public)},
				{CKA_PRIVATE, &pfalse, sizeof(pfalse)}  // private object
			};
			BYTE    numAttributes_public = 5;

			// Open public session
			TS_ASSERT(m_pkcs11Mngr.C_OpenSession(slotID, CKF_RW_SESSION | CKF_SERIAL_SESSION, NULL_PTR, NULL_PTR, &hPublicSession) == CKR_OK);
			// Open private session
			TS_ASSERT(m_pkcs11Mngr.C_OpenSession(slotID, CKF_RW_SESSION | CKF_SERIAL_SESSION, NULL_PTR, NULL_PTR, &hPrivateSession) == CKR_OK);
			// Login user
			TS_ASSERT(m_pkcs11Mngr.C_Login(hPrivateSession, CKU_USER, (CK_CHAR*)TEST_PIN, strlen(TEST_PIN)) == CKR_OK);

			// Create new public object 
			TS_ASSERT(m_pkcs11Mngr.C_CreateObject(hPublicSession, dataTemplate_public, numAttributes_public, &hObject[0]) == CKR_OK);
			// Create new private object - should succed
			TS_ASSERT(m_pkcs11Mngr.C_CreateObject(hPrivateSession, dataTemplate_private, numAttributes_private, &hObject[1]) == CKR_OK);

			// Try to find public object in public session - should succedd
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsInit(hPublicSession, dataTemplate_public, numAttributes_public) == CKR_OK); 
			TS_ASSERT(m_pkcs11Mngr.C_FindObjects(hPublicSession, hObject, NUM_OBJECTS, &ulObjectCount) == CKR_OK);	
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsFinal(hPublicSession) == CKR_OK); 
			// One object should be present
			TS_ASSERT(ulObjectCount == 1);

			// Try to find public object in private session - should succedd
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsInit(hPrivateSession, dataTemplate_public, numAttributes_public) == CKR_OK); 
			TS_ASSERT(m_pkcs11Mngr.C_FindObjects(hPrivateSession, hObject, NUM_OBJECTS, &ulObjectCount) == CKR_OK);	
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsFinal(hPrivateSession) == CKR_OK); 
			// One object should be present
			TS_ASSERT(ulObjectCount == 1);

			// Try to find private object in public session - should fail
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsInit(hPublicSession, dataTemplate_private, numAttributes_private) == CKR_OK); 
			TS_ASSERT(m_pkcs11Mngr.C_FindObjects(hPublicSession, hObject, NUM_OBJECTS, &ulObjectCount) == CKR_OK);	
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsFinal(hPublicSession) == CKR_OK); 
			// No object should be visible
			TS_ASSERT(ulObjectCount == 0);

			// Try to find private object in private session - should succedd
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsInit(hPrivateSession, dataTemplate_private, numAttributes_private) == CKR_OK); 
			TS_ASSERT(m_pkcs11Mngr.C_FindObjects(hPrivateSession, hObject, NUM_OBJECTS, &ulObjectCount) == CKR_OK);	
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsFinal(hPrivateSession) == CKR_OK); 
			// One object should be present
			TS_ASSERT(ulObjectCount == 1);


			m_pkcs11Mngr.C_CloseAllSessions(slotID);
		}
/**/
	}

    void testInternalObjectsVisibility(void) {
		int             status = STAT_OK;

		CK_SLOT_ID	slotID = -100;
		CK_SLOT_INFO    slotInfo;
		CK_TOKEN_INFO   tokenInfo;
		TS_ASSERT((status = getTestTokenSession(&slotID, &slotInfo, &tokenInfo)) == CKR_OK);
		if (status == CKR_OK) {
			CK_SESSION_HANDLE	hPublicSession;

			CK_OBJECT_HANDLE hObject[NUM_OBJECTS];
			CK_ULONG ulObjectCount;
			// TEMPLATE to use for creation/searching object on token
			CK_OBJECT_CLASS dataClass = CKO_DATA;  //type of data in object
			CK_CHAR label[] = {"_CAO_Test4"};	//label of data object
			CK_ATTRIBUTE dataTemplate[] = {
				{CKA_CLASS, &dataClass, sizeof(dataClass)},
				{CKA_LABEL, label, sizeof(label)}				
			};


			// Open public session
			TS_ASSERT(m_pkcs11Mngr.C_OpenSession(slotID, CKF_RW_SESSION | CKF_SERIAL_SESSION, NULL_PTR, NULL_PTR, &hPublicSession) == CKR_OK);

			// Create new public object 
			TS_ASSERT(m_pkcs11Mngr.C_CreateObject(hPublicSession, dataTemplate, 2, &hObject[0]) == CKR_OK);

			//
			// Enable search for internal objects  
			//
			m_pkcs11Mngr.C_CloseAllSessions(slotID);
			m_pkcs11Mngr.CAM_EnableInternalObjectsVisibility(slotID, TRUE);

			// Open public session
			TS_ASSERT(m_pkcs11Mngr.C_OpenSession(slotID, CKF_RW_SESSION | CKF_SERIAL_SESSION, NULL_PTR, NULL_PTR, &hPublicSession) == CKR_OK);
			// Try to find created object in public session - should succedd as we have seraching for private objects enabled
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsInit(hPublicSession, dataTemplate, 2) == CKR_OK); 
			TS_ASSERT(m_pkcs11Mngr.C_FindObjects(hPublicSession, hObject, NUM_OBJECTS, &ulObjectCount) == CKR_OK);	
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsFinal(hPublicSession) == CKR_OK); 
			// On object should be present
			TS_ASSERT(ulObjectCount == 1);


			//
			// Disable search for internal objects  
			//
			m_pkcs11Mngr.C_CloseAllSessions(slotID);
			m_pkcs11Mngr.CAM_EnableInternalObjectsVisibility(slotID, FALSE);

			// Open public session
			TS_ASSERT(m_pkcs11Mngr.C_OpenSession(slotID, CKF_RW_SESSION | CKF_SERIAL_SESSION, NULL_PTR, NULL_PTR, &hPublicSession) == CKR_OK);
			// Try to find created object in public session - should succedd as we have seraching for private objects enabled
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsInit(hPublicSession, dataTemplate, 2) == CKR_OK); 
			TS_ASSERT(m_pkcs11Mngr.C_FindObjects(hPublicSession, hObject, NUM_OBJECTS, &ulObjectCount) == CKR_OK);	
			TS_ASSERT(m_pkcs11Mngr.C_FindObjectsFinal(hPublicSession) == CKR_OK); 
			// No object should be returned
			TS_ASSERT(ulObjectCount == 0);

			m_pkcs11Mngr.C_CloseAllSessions(slotID);
		}
	}


};



#endif
